package PScm::Env;

use strict;
use warnings;
use base qw(PScm);
use PScm::Continuation;

sub new {
    my ($class, %bindings) = @_;

    $class = ref($class) || $class;
    bless { bindings => {%bindings}, parent => 0 }, $class;
}

sub Extend {
    my ($self, $ra_symbols, $ra_values, $cont) = @_;

    PScm::Expr::List->new(@$ra_values)->map_eval(
        $self,
        cont {
            my ($ra_evaluated_values) = @_;
            $cont->Cont(
                $self->ExtendUnevaluated(
                    $ra_symbols, $ra_evaluated_values
                )
            );
        }
    );
}

sub ExtendUnevaluated {
    my ($self, $ra_symbols, $ra_values) = @_;

    my %newbindings;
    @newbindings{ map { $_->value } @$ra_symbols } = @$ra_values;
    my $newenv = $self->new(%newbindings);
    $newenv->{parent} = $self;
    return $newenv;
}

sub ExtendRecursively {
    my ($self, $ra_symbols, $ra_values, $cont) = @_;

    my $newenv = $self->ExtendUnevaluated($ra_symbols, $ra_values);
    $newenv->_eval_values($cont);
}

sub ExtendIteratively {
    my ($self, $ra_symbols, $ra_values, $cont) = @_;
    my @symbols = @$ra_symbols;
    my @values  = @$ra_values;
    if (@symbols) {
        my $symbol = shift @symbols;
        my $value  = shift @values;
        $self->Extend(
            [$symbol],
            [$value],
            cont {
                my ($extended) = @_;
                $extended->ExtendIteratively([@symbols], [@values],
                    $cont);
            }
        );
    } else {
        $cont->Cont($self);
    }
}

sub _eval_values {
    my ($self, $cont) = @_;
    $self->_map_bindings([keys %{ $self->{bindings} }], $cont);
}

sub _map_bindings {
    my ($self, $ra_keys, $cont) = @_;
    my (@keys) = @$ra_keys;
    if (@keys) {
        my $firstkey = shift @keys;
        $self->{bindings}{$firstkey}->Eval(
            $self,
            cont {
                my ($value) = @_;
                $self->{bindings}{$firstkey} = $value;
                $self->_map_bindings([@keys], $cont);
            }
        );
    } else {
        $cont->Cont($self);
    }
}

sub LookUp {
    my ($self, $symbol) = @_;

    if (defined(my $ref = $self->_lookup_ref($symbol))) {
        return $$ref;
    } else {
        die "no binding for @{[$symbol->value]}",
          " in @{[ref($self)]}\n";
    }
}

sub Assign {
    my ($self, $symbol, $value) = @_;

    if (defined(my $ref = $self->_lookup_ref($symbol))) {
        $$ref = $value;
    } else {
        die "no binding for @{[$symbol->value]}",
          " in @{[ref($self)]}\n";
    }
}

sub _lookup_ref {
    my ($self, $symbol) = @_;

    if (exists($self->{bindings}{ $symbol->value })) {
        return \$self->{bindings}{ $symbol->value };
    } elsif ($self->{parent}) {
        return $self->{parent}->_lookup_ref($symbol);
    } else {
        return undef;
    }
}

1;
